@@ -1,28 +1,16 @@
 /*
- *   Copyright 2011 Marco Martin <mart@kde.org>
- *   Copyright 2014 Aleix Pol Gonzalez <aleixpol@blue-systems.com>
+ *  SPDX-FileCopyrightText: 2011 Marco Martin <mart@kde.org>
+ *  SPDX-FileCopyrightText: 2014 Aleix Pol Gonzalez <aleixpol@blue-systems.com>
  *
- *   This program is free software; you can redistribute it and/or modify
- *   it under the terms of the GNU Library General Public License as
- *   published by the Free Software Foundation; either version 2, or
- *   (at your option) any later version.
- *
- *   This program is distributed in the hope that it will be useful,
- *   but WITHOUT ANY WARRANTY; without even the implied warranty of
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *   GNU General Public License for more details
- *
- *   You should have received a copy of the GNU Library General Public
- *   License along with this program; if not, write to the
- *   Free Software Foundation, Inc.,
- *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+ *  SPDX-License-Identifier: LGPL-2.0-or-later
  */
 
 #include "icon.h"
 #include "libkirigami/platformtheme.h"
+#include "scenegraph/managedtexturenode.h"
 
 #include <QSGSimpleTextureNode>
-#include <qquickwindow.h>
+#include <QQuickWindow>
 #include <QIcon>
 #include <QBitmap>
 #include <QSGTexture>
@@ -35,111 +23,24 @@
 #include <QPainter>
 #include <QScreen>
 
-class ManagedTextureNode : public QSGSimpleTextureNode
-{
-Q_DISABLE_COPY(ManagedTextureNode)
-public:
-    ManagedTextureNode();
 
-    void setTexture(QSharedPointer<QSGTexture> texture);
-
-private:
-    QSharedPointer<QSGTexture> m_texture;
-};
-
-ManagedTextureNode::ManagedTextureNode()
-{}
-
-void ManagedTextureNode::setTexture(QSharedPointer<QSGTexture> texture)
-{
-    m_texture = texture;
-    QSGSimpleTextureNode::setTexture(texture.data());
-}
-
-typedef QHash<qint64, QHash<QWindow*, QWeakPointer<QSGTexture> > > TexturesCache;
-
-struct ImageTexturesCachePrivate
-{
-    TexturesCache cache;
-};
-
-class ImageTexturesCache
-{
-public:
-    ImageTexturesCache();
-    ~ImageTexturesCache();
-
-    /**
-     * @returns the texture for a given @p window and @p image.
-     *
-     * If an @p image id is the same as one already provided before, we won't create
-     * a new texture and return a shared pointer to the existing texture.
-     */
-    QSharedPointer<QSGTexture> loadTexture(QQuickWindow *window, const QImage &image, QQuickWindow::CreateTextureOptions options);
-
-    QSharedPointer<QSGTexture> loadTexture(QQuickWindow *window, const QImage &image);
-
-
-private:
-    QScopedPointer<ImageTexturesCachePrivate> d;
-};
-
-
-ImageTexturesCache::ImageTexturesCache()
-    : d(new ImageTexturesCachePrivate)
-{
-}
-
-ImageTexturesCache::~ImageTexturesCache()
-{
-}
-
-QSharedPointer<QSGTexture> ImageTexturesCache::loadTexture(QQuickWindow *window, const QImage &image, QQuickWindow::CreateTextureOptions options)
-{
-    qint64 id = image.cacheKey();
-    QSharedPointer<QSGTexture> texture = d->cache.value(id).value(window).toStrongRef();
-
-    if (!texture) {
-        auto cleanAndDelete = [this, window, id](QSGTexture* texture) {
-            QHash<QWindow*, QWeakPointer<QSGTexture> >& textures = (d->cache)[id];
-            textures.remove(window);
-            if (textures.isEmpty())
-                d->cache.remove(id);
-            delete texture;
-        };
-        texture = QSharedPointer<QSGTexture>(window->createTextureFromImage(image, options), cleanAndDelete);
-        (d->cache)[id][window] = texture.toWeakRef();
-    }
-
-    //if we have a cache in an atlas but our request cannot use an atlassed texture
-    //create a new texture and use that
-    //don't use removedFromAtlas() as that requires keeping a reference to the non atlased version
-    if (!(options & QQuickWindow::TextureCanUseAtlas) && texture->isAtlasTexture()) {
-        texture = QSharedPointer<QSGTexture>(window->createTextureFromImage(image, options));
-    }
-
-    return texture;
-}
-
-QSharedPointer<QSGTexture> ImageTexturesCache::loadTexture(QQuickWindow *window, const QImage &image)
-{
-    return loadTexture(window, image, {});
-}
 
 Q_GLOBAL_STATIC(ImageTexturesCache, s_iconImageCache)
 
 Icon::Icon(QQuickItem *parent)
     : QQuickItem(parent),
-      m_smooth(false),
       m_changed(false),
       m_active(false),
       m_selected(false),
       m_isMask(false)
 {
     setFlag(ItemHasContents, true);
+    // Using 32 because Icon used to redefine implicitWidth and implicitHeight and hardcode them to 32
+    setImplicitSize(32, 32);
     //FIXME: not necessary anymore
     connect(qApp, &QGuiApplication::paletteChanged, this, &QQuickItem::polish);
     connect(this, &QQuickItem::enabledChanged, this, &QQuickItem::polish);
+    connect(this, &QQuickItem::smoothChanged, this, &QQuickItem::polish);
 }
 
 
@@ -175,6 +76,7 @@ void Icon::setSource(const QVariant &icon)
         m_networkReply->close();
     }
     m_loadedImage = QImage();
+    setStatus(Loading);
 
     polish();
     emit sourceChanged();
@@ -203,6 +105,12 @@ bool Icon::active() const
 
 bool Icon::valid() const
 {
+    // TODO: should this be return m_status == Ready?
+    // Consider an empty URL invalid, even though isNull() will say false
+    if (m_source.canConvert<QUrl>() && m_source.toUrl().isEmpty()) {
+        return false;
+    }
+
     return !m_source.isNull();
 }
 
@@ -254,35 +162,9 @@ QColor Icon::color() const
     return m_color;
 }
 
-
-int Icon::implicitWidth() const
-{
-    return 32;
-}
-
-int Icon::implicitHeight() const
-{
-    return 32;
-}
-
-void Icon::setSmooth(const bool smooth)
-{
-    if (smooth == m_smooth) {
-        return;
-    }
-    m_smooth = smooth;
-    polish();
-    emit smoothChanged();
-}
-
-bool Icon::smooth() const
-{
-    return m_smooth;
-}
-
 QSGNode* Icon::updatePaintNode(QSGNode* node, QQuickItem::UpdatePaintNodeData* /*data*/)
 {
-    if (m_source.isNull()) {
+    if (m_source.isNull() || qFuzzyIsNull(width()) || qFuzzyIsNull(height())) {
         delete node;
         return Q_NULLPTR;
     }
@@ -310,7 +192,7 @@ QSGNode* Icon::updatePaintNode(QSGNode* node, QQuickItem::UpdatePaintNodeData* /
         }
         mNode->setRect(nodeRect);
         node = mNode;
-        if (m_smooth) {
+        if (smooth()) {
             mNode->setFiltering(QSGTexture::Linear);
         }
         m_changed = false;
@@ -364,8 +246,7 @@ void Icon::handleFinished(QNetworkReply* reply)
         qWarning() << "received broken image" << reply->url();
 
         // broken image from data, inform the user of this with some useful broken-image thing...
-        const QSize size = QSize(width(), height()) * (window() ? window()->devicePixelRatio() : qApp->devicePixelRatio());
-        m_loadedImage = QIcon::fromTheme(m_fallback).pixmap(size, iconMode(), QIcon::On).toImage();
+        m_loadedImage = QIcon::fromTheme(m_fallback).pixmap(window(), QSize(width(), height()), iconMode(), QIcon::On).toImage();
     }
     polish();
 }
@@ -375,6 +256,7 @@ void Icon::updatePolish()
     QQuickItem::updatePolish();
 
     if (m_source.isNull()) {
+        updatePaintedGeometry();
         return;
     }
 
@@ -394,7 +276,7 @@ void Icon::updatePolish()
             m_icon = m_source.value<QBitmap>().toImage();
             break;
         case QVariant::Icon:
-            m_icon = m_source.value<QIcon>().pixmap(size, iconMode(), QIcon::On).toImage();
+            m_icon = m_source.value<QIcon>().pixmap(window(), itemSize, iconMode(), QIcon::On).toImage();
             break;
         case QVariant::Url:
         case QVariant::String:
@@ -426,6 +308,7 @@ void Icon::updatePolish()
         }
     }
     m_changed = true;
+    updatePaintedGeometry();
     update();
 }
 
@@ -435,6 +318,7 @@ QImage Icon::findIcon(const QSize &size)
     QString iconSource = m_source.toString();
 
     if (iconSource.startsWith(QLatin1String("image://"))) {
+        const auto multiplier = QCoreApplication::instance()->testAttribute(Qt::AA_UseHighDpiPixmaps) ? (window() ? window()->devicePixelRatio() : qGuiApp->devicePixelRatio()) : 1;
         QUrl iconUrl(iconSource);
         QString iconProviderId = iconUrl.host();
         QString iconId = iconUrl.path();
@@ -451,20 +335,63 @@ QImage Icon::findIcon(const QSize &size)
             return img;
         switch(imageProvider->imageType()){
         case QQmlImageProviderBase::Image:
-            img = imageProvider->requestImage(iconId, &actualSize, size);
+            img = imageProvider->requestImage(iconId, &actualSize, size * multiplier);
             break;
         case QQmlImageProviderBase::Pixmap:
-            img = imageProvider->requestPixmap(iconId, &actualSize, size).toImage();
+            img = imageProvider->requestPixmap(iconId, &actualSize, size * multiplier).toImage();
             break;
-        case QQmlImageProviderBase::Texture:
-        case QQmlImageProviderBase::Invalid:
         case QQmlImageProviderBase::ImageResponse:
+        {
+            if (!m_loadedImage.isNull()) {
+                setStatus(Ready);
+                return m_loadedImage.scaled(size, Qt::KeepAspectRatio, smooth() ? Qt::SmoothTransformation : Qt::FastTransformation );
+            }
+            QQuickAsyncImageProvider *provider = dynamic_cast<QQuickAsyncImageProvider*>(imageProvider);
+            auto response = provider->requestImageResponse(iconId, size * multiplier);
+            connect(response, &QQuickImageResponse::finished, this, [iconId, response, this](){
+                if (response->errorString().isEmpty()) {
+                    QQuickTextureFactory *textureFactory = response->textureFactory();
+                    if (textureFactory) {
+                        m_loadedImage = textureFactory->image();
+                    }
+                    if (m_loadedImage.isNull()) {
+                        // broken image from data, inform the user of this with some useful broken-image thing...
+                        m_loadedImage = QIcon::fromTheme(m_fallback).pixmap(window(), QSize(width(), height()), iconMode(), QIcon::On).toImage();
+                        setStatus(Error);
+                    } else {
+                        setStatus(Ready);
+                    }
+                    polish();
+                }
+            });
+            // Temporary icon while we wait for the real image to load...
+            img = QIcon::fromTheme(m_placeholder).pixmap(window(), size, iconMode(), QIcon::On).toImage();
+            break;
+        }
+        case QQmlImageProviderBase::Texture:
+        {
+            QQuickTextureFactory *textureFactory = imageProvider->requestTexture(iconId, &actualSize, size * multiplier);
+            if (textureFactory) {
+                img = textureFactory->image();
+            }
+            if (img.isNull()) {
+                // broken image from data, or the texture factory wasn't healthy, inform the user of this with some useful broken-image thing...
+                img = QIcon::fromTheme(m_fallback).pixmap(window(), QSize(width(), height()), iconMode(), QIcon::On).toImage();
+                setStatus(Error);
+            } else {
+                setStatus(Ready);
+            }
+            break;
+        }
+        case QQmlImageProviderBase::Invalid:
             //will have to investigate this more
+            setStatus(Error);
             break;
         }
     } else if(iconSource.startsWith(QLatin1String("http://")) || iconSource.startsWith(QLatin1String("https://"))) {
         if(!m_loadedImage.isNull()) {
-            return m_loadedImage.scaled(size, Qt::KeepAspectRatio, m_smooth ? Qt::SmoothTransformation : Qt::FastTransformation );
+            setStatus(Ready);
+            return m_loadedImage.scaled(size, Qt::KeepAspectRatio, smooth() ? Qt::SmoothTransformation : Qt::FastTransformation );
         }
         const auto url = m_source.toUrl();
         QQmlEngine* engine = qmlEngine(this);
@@ -476,7 +403,7 @@ QImage Icon::findIcon(const QSize &size)
             connect(m_networkReply.data(), &QNetworkReply::finished, this, [this](){ handleFinished(m_networkReply); });
         }
         // Temporary icon while we wait for the real image to load...
-        img = QIcon::fromTheme(QStringLiteral("image-x-icon")).pixmap(size, iconMode(), QIcon::On).toImage();
+        img = QIcon::fromTheme(m_placeholder).pixmap(window(), size, iconMode(), QIcon::On).toImage();
     } else {
         if (iconSource.startsWith(QLatin1String("qrc:/"))) {
             iconSource = iconSource.mid(3);
@@ -494,8 +421,9 @@ QImage Icon::findIcon(const QSize &size)
             }
         }
         if (!icon.isNull()) {
-            img = icon.pixmap(size, iconMode(), QIcon::On).toImage();
+            img = icon.pixmap(window(), size, iconMode(), QIcon::On).toImage();
 
+            setStatus(Ready);
             /*const QColor tintColor = !m_color.isValid() || m_color == Qt::transparent ? (m_selected ? m_theme->highlightedTextColor() : m_theme->textColor()) : m_color;
 
             if (m_isMask || icon.isMask() || iconSource.endsWith(QLatin1String("-symbolic")) || iconSource.endsWith(QLatin1String("-symbolic-rtl")) || iconSource.endsWith(QLatin1String("-symbolic-ltr")) || guessMonochrome(img)) {
@@ -506,6 +434,11 @@ QImage Icon::findIcon(const QSize &size)
             }*/
         }
     }
+
+    if (!iconSource.isEmpty() && img.isNull()) {
+        setStatus(Error);
+        img = QIcon::fromTheme(m_fallback).pixmap(window(), size, iconMode(), QIcon::On).toImage();
+    }
     return img;
 }
 
@@ -568,10 +501,9 @@ bool Icon::guessMonochrome(const QImage &img)
 
     QMultiMap<int, int> reverseDist;
     auto it = dist.constBegin();
-    std::vector<qreal> probabilities(dist.size());
     qreal entropy = 0;
     while (it != dist.constEnd()) {
-        reverseDist.insertMulti(it.value(), it.key());
+        reverseDist.insert(it.value(), it.key());
         qreal probability = qreal(it.value()) / qreal(img.size().width() * img.size().height() - transparentPixels);
         entropy -= probability * log(probability) / log(255);
         ++it;
@@ -595,4 +527,68 @@ void Icon::setFallback(const QString& fallback)
     }
 }
 
+QString Icon::placeholder() const
+{
+    return m_placeholder;
+}
+
+void Icon::setPlaceholder(const QString& placeholder)
+{
+    if (m_placeholder != placeholder) {
+        m_placeholder = placeholder;
+        Q_EMIT placeholderChanged(placeholder);
+    }
+}
+
+void Icon::setStatus(Status status)
+{
+    if (status == m_status) {
+        return;
+    }
+
+    m_status = status;
+    emit statusChanged();
+}
+
+Icon::Status Icon::status() const
+{
+    return m_status;
+}
+
+qreal Icon::paintedWidth() const
+{
+    return m_paintedWidth;
+}
+
+qreal Icon::paintedHeight() const
+{
+    return m_paintedHeight;
+}
+
+void Icon::updatePaintedGeometry()
+{
+    qreal newWidth = 0.0;
+    qreal newHeight = 0.0;
+    if (!m_icon.width() || !m_icon.height()) {
+        newWidth = newHeight = 0.0;
+    } else {
+        const qreal w = widthValid() ? width() : m_icon.width();
+        const qreal widthScale = w / m_icon.width();
+        const qreal h = heightValid() ? height() : m_icon.height();
+        const qreal heightScale = h / m_icon.height();
+        if (widthScale <= heightScale) {
+            newWidth = w;
+            newHeight = widthScale * m_icon.height();
+        } else if (heightScale < widthScale) {
+            newWidth = heightScale * m_icon.width();
+            newHeight = h;
+        }
+    }
+    if (newWidth != m_paintedWidth || newHeight != m_paintedHeight) {
+        m_paintedWidth = newWidth;
+        m_paintedHeight = newHeight;
+        Q_EMIT paintedAreaChanged();
+    }
+}
+
 #include "moc_icon.cpp"
